%{
/* actionparser.y
 *
 * Yacc script for parsing an action
 */
#include "sim.h"
#include "server.h"

static pnode *new_pnode(int tp);
static pnode *create_binop(int tp, pnode *l, pnode *r);
static void write_variable(int fd, int tl, variable *v);
static void write_method(int fd, int tl, method *m);
static void write_ptree(int fd, int tl, pnode *n);

extern int parse_debug;
%}

%union {
	double num;
	char string[256];
	pnode *pn;
	int integer;
	}
%token <num> NUM		/* numeric constant */
%token <string> STRING		/* quoted string */
%token <string> WORD		/* unquoted string (ie-var name) */
%nonassoc LOWER_THAN_ELSE
%nonassoc END IF ELSE FOR WHILE DO ACTION METHOD STR DBL LOCAL STATIC
%right '=' PLUSEQ MINUSEQ TIMESEQ DIVEQ
%left ','
%nonassoc '?' ':'
%left AND
%left OR
%nonassoc '<' '>' EQ NEQ GEQ LEQ
%left '-' '+'
%left '*' '/' '%' '&'
%nonassoc UMINUS UPLUS UNOT PLUSPLUS MINUSMINUS SIZEOF TYPEOF
%type <pn> statement statements expr lval arglist arraylist expr_empty opt_val
%type <string> word_or_string
%type <integer> storage
%type <num> array_size
%%
input		:	END
			{ return -1;
			}
		|	action
			{ return 0;
			}
		;

action		:	ACTION word_or_string '{'
			{ na = create_action();
			  strcpy(na->name,$2);
			}
			actionattribs '}'
		;

word_or_string	:	WORD
		|	STRING
		;

actionattribs	:	actionattribs actionattrib
		|
		;

actionattrib	:	variable
			{ na->var[na->varc++] = nv;
			}
		|	method
		;


/* variable declaration and possible assignment with initial value
 */
variable	:	str_dec 
		|	dbl_dec
		|	str_array_dec
		|	dbl_array_dec
		;

str_dec		:	storage STR WORD opt_val ';'
			{ nv = create_variable($3,vt_str,$1);
			  nv->val.str = scalloc(1,1);
			  nv->init = $4;
			}
		;

storage		:	LOCAL
			{ $$ = vst_local; }
		|	STATIC
			{ $$ = vst_static; }
		|	{ $$ = vst_local; }
		;

opt_val		:	'=' expr
			{ $$ = $2; }
		|	{ $$ = NULL; }
		;

dbl_dec		:	storage DBL WORD opt_val ';'
			{ nv = create_variable($3,vt_dbl,$1);
			  nv->val.dbl = 0.0;
			  nv->init = $4;
			}
		;

str_array_dec	:	storage STR '[' array_size ']' WORD opt_val ';'
			{ int i,size=$4;
			  if (size < 0) {
			    yyerror("Invalid array size");
			    return 1;
			    }
			  nv = create_variable($6,vt_strarray,$1);
			  nv->val.arr.a = scalloc(size,sizeof(variable));
			  nv->val.arr.n = size;
			  nv->val.arr.ni = size;
			  nv->init = $7;
			  for(i=0; i<size; i++) {
			    variable *sub = &nv->val.arr.a[i];
			    sub->type = vt_str;
			    sub->storage = $1;
			    sub->val.str = scalloc(1,1);
			    }
			}
		;

array_size	:	NUM
			{ $$ = $1; }
		|	{ $$ = 0; }
		;

dbl_array_dec	:	storage DBL '[' array_size ']' WORD opt_val ';'
			{ int i, size=$4;
			  if (size < 0) {
			    yyerror("Invalid array size");
			    return 1;
			    }
			  nv = create_variable($6,vt_dblarray,$1);
			  nv->val.arr.a = scalloc(size,sizeof(variable));
			  nv->val.arr.n = size;
			  nv->val.arr.ni = size;
			  nv->init = $7;
			  for(i=0; i<size; i++) {
			    variable *sub = &nv->val.arr.a[i];
			    sub->type = vt_dbl;
			    sub->storage = $1;
			    sub->val.dbl = 0.0;
			    }
			}
		;

/* a method contains local variable declarations followed by code
 */
method		:	METHOD
			{ nm = create_method();
			}
			method_names '{' method_vars statements '}'
			{ pnode *nn = new_pnode(pn_cmds);
			  nn->val.nblock.st = $6;
			  nm->ptree = nn;
			  na->met[na->metc++] = nm;
			}
		;

method_names	:	method_names method_name
		|	method_name
		;

method_name	:	WORD
			{ nm->name[nm->namec++] = strdupe($1);
			}
		;

method_vars	:	method_vars method_var
		|
		;

method_var	:	variable
			{ nm->var[nm->varc++] = nv;
			}
		;

statements	:	statement statements
			{ $1->next = $2;
			  $$ = $1;
			}
		|	{ $$ = NULL; }
		;

statement	:	IF '(' expr ')' statement %prec LOWER_THAN_ELSE
			{ pnode *nn = new_pnode(pn_if);
			  nn->val.nif.cond = $3;
			  nn->val.nif.tr = $5;
			  nn->val.nif.fl = NULL;
			  $$ = nn;
			}
		|	IF '(' expr ')' statement ELSE statement
			{ pnode *nn = new_pnode(pn_if);
			  nn->val.nif.cond = $3;
			  nn->val.nif.tr = $5;
			  nn->val.nif.fl = $7;
			  $$ = nn;
			}
		|	WHILE '(' expr ')'
				statement
			{ pnode *nn = new_pnode(pn_while);
			  nn->val.nwhile.cond = $3;
			  nn->val.nwhile.blk = $5;
			  $$ = nn;
			}
		|	FOR '(' expr_empty ';' expr_empty ';' expr_empty ')'
				statement
			{ pnode *nn = new_pnode(pn_for);
			  nn->val.nfor.init = $3;
			  nn->val.nfor.cond = $5;
			  nn->val.nfor.blk  = $9;
			  nn->val.nfor.inc  = $7;
			  $$ = nn;
			}
		|	DO
				statement
			WHILE '(' expr ')' ';'
			{ pnode *nn = new_pnode(pn_do);
			  nn->val.ndo.cond = $5;
			  nn->val.ndo.blk  = $2;
			  $$ = nn;
			}
		|	'{' statements '}'
			{ pnode *nn = new_pnode(pn_block);
			  nn->val.nblock.st = $2;
			  $$ = nn;
			}
		|	expr ';'
			{ pnode *nn = new_pnode(pn_estat);
			  nn->val.nestat.exp = $1;
			  $$ = nn;
			}
		|	';'
			{ pnode *nn = new_pnode(pn_null);
			  $$ = nn;
			}
		;

expr		:	expr EQ expr
			{ $$ = create_binop(pn_equals,$1,$3); }
		|	expr NEQ expr
			{ $$ = create_binop(pn_notequals,$1,$3); }
		|	expr '+' expr
			{ $$ = create_binop(pn_add,$1,$3); }
		|	expr '-' expr
			{ $$ = create_binop(pn_subtract,$1,$3); }
		|	expr '*' expr
			{ $$ = create_binop(pn_mult,$1,$3); }
		|	expr '/' expr
			{ $$ = create_binop(pn_divide,$1,$3); }
		|	expr '%' expr
			{ $$ = create_binop(pn_mod,$1,$3); }
		|	expr '>' expr
			{ $$ = create_binop(pn_greater,$1,$3); }
		|	expr '<' expr
			{ $$ = create_binop(pn_less,$1,$3); }
		|	expr GEQ expr
			{ $$ = create_binop(pn_grequals,$1,$3); }
		|	expr LEQ expr
			{ $$ = create_binop(pn_lsequals,$1,$3); }
		|	expr OR expr
			{ $$ = create_binop(pn_or,$1,$3); }
		|	expr AND expr
			{ $$ = create_binop(pn_and,$1,$3); }
		|	'!' expr %prec UNOT
			{ pnode *nn = new_pnode(pn_unot);
			  nn->val.nunop.r = $2;
			  $$ = nn;
			}
		|	'-' expr %prec UMINUS
			{ pnode *nn = new_pnode(pn_uminus);
			  nn->val.nunop.r = $2;
			  $$ = nn;
			}
		|	'+' expr %prec UPLUS
			{ pnode *nn = new_pnode(pn_uplus);
			  nn->val.nunop.r = $2;
			  $$ = nn;
			}
		|	lval
			{ pnode *nn = new_pnode(pn_var);
			  nn->val.nvar.var = $1;
			  $$ = nn;
			}
		|	lval '=' expr
			{ pnode *nn = new_pnode(pn_assign);
			  nn->val.nassign.lv = $1;
			  nn->val.nassign.val = $3;
			  $$ = nn;
			}
		|	lval PLUSEQ expr
			{ pnode *nn = new_pnode(pn_addeq);
			  nn->val.nlvalop.lv = $1;
			  nn->val.nlvalop.val = $3;
			  $$ = nn;
			}
		|	lval MINUSEQ expr
			{ pnode *nn = new_pnode(pn_subtracteq);
			  nn->val.nlvalop.lv = $1;
			  nn->val.nlvalop.val = $3;
			  $$ = nn;
			}
		|	lval TIMESEQ expr
			{ pnode *nn = new_pnode(pn_multeq);
			  nn->val.nlvalop.lv = $1;
			  nn->val.nlvalop.val = $3;
			  $$ = nn;
			}
		|	lval DIVEQ expr
			{ pnode *nn = new_pnode(pn_divideeq);
			  nn->val.nlvalop.lv = $1;
			  nn->val.nlvalop.val = $3;
			  $$ = nn;
			}
		|	'(' expr ')'
			{ $$ = $2;
			}
		|	WORD '(' arglist ')'
			{ pnode *nn = new_pnode(pn_fcall);
			  if (!find_runtime_func($1)) {
			    yyerror("Function not found");
			    return 1;
			    }
			  nn->val.nfcall.name = strdupe($1);
			  nn->val.nfcall.ar = $3;
			  $$ = nn;
			}
		|	NUM			/* constant number */
			{ pnode *nn = new_pnode(pn_const);
			  nn->val.nconst.n = $1;
			  $$ = nn;
			}
		|	expr '&' expr
			{ pnode *nn = new_pnode(pn_concat);
			  nn->val.nconcat.l = $1;
			  nn->val.nconcat.r = $3;
			  $$ = nn;
			}
		|	STRING			/* constant string */
			{ pnode *nn = new_pnode(pn_sconst);
			  nn->val.nsconst.str = strdupe($1);
			  $$ = nn;
			}
		|	'[' arraylist ']'	/* constant array */
			{ pnode *nn = new_pnode(pn_aconst);
			  nn->val.naconst.l = $2;
			  $$ = nn;
			}
/*
		|	expr '[' expr ']'
			{ pnode *nn = new_pnode(pn_aindex);
			  nn->val.naindex.arr = $1;
			  nn->val.naindex.ind = $3;
			  $$ = nn;
			}
*/
		|	PLUSPLUS lval
			{ pnode *nn = new_pnode(pn_preinc);
			  nn->val.nincdec.v = $2;
			  $$ = nn;
			}
		|	lval PLUSPLUS
			{ pnode *nn = new_pnode(pn_postinc);
			  nn->val.nincdec.v = $1;
			  $$ = nn;
			}
		|	MINUSMINUS lval
			{ pnode *nn = new_pnode(pn_predec);
			  nn->val.nincdec.v = $2;
			  $$ = nn;
			}
		|	lval MINUSMINUS
			{ pnode *nn = new_pnode(pn_postdec);
			  nn->val.nincdec.v = $1;
			  $$ = nn;
			}
		|	SIZEOF '(' expr ')'
			{ pnode *nn = new_pnode(pn_sizeof);
			  nn->val.nsizeof.a = $3;
			  $$ = nn;
			}
		|	TYPEOF '(' expr ')'
			{ pnode *nn = new_pnode(pn_typeof);
			  nn->val.ntypeof.e = $3;
			  $$ = nn;
			}
		;

expr_empty	:	expr { $$ = $1; }
		|	{ pnode *nn = new_pnode(pn_const);
			  nn->val.nconst.n = 1;
			  $$ = nn;
			}
		;


lval		:	WORD
			{ pnode *nn = new_pnode(pn_lval);
			  variable *loc = lookup_variable($1,na,nm);

			  if (!loc) {
			    yyerror("Variable not defined");
			    return 1;
			    }
			  nn->val.nlval.loc = loc;
			  $$ = nn;
			}
		|	lval '[' expr ']'
			{ pnode *nn = new_pnode(pn_aindex);
			  nn->val.naindex.arr = $1;
			  nn->val.naindex.ind = $3;
			  $$ = nn;
			}
		;

arglist		:	expr ',' arglist
			{ $1->next = $3;
			  $$ = $1;
			}
		|	expr
			{ $$ = $1; }
		|	{ $$ = NULL; }
		;

arraylist	:	expr ';' arraylist
			{ $1->next = $3;
			  $$ = $1;
			}
		|	expr
			{ $$ = $1; }
		|	{ $$ = NULL; }
		;

%%
static action *na;		/* action being read */
static method *nm;		/* method being read */
static variable *nv;		/* variable being read */
int fd;				/* fd data is being read from */
char lineno;			/* lines read in input */
static int currch;

static reserved rwtab[] = {	"end",END,"if",IF,"else",ELSE,"for",FOR,
				"while",WHILE,"action",ACTION,"method",METHOD,
				"str",STR,"dbl",DBL,"do",DO,"local",LOCAL,
				"static",STATIC,"sizeof",SIZEOF,
				"typeof",TYPEOF };

/* nextchar
 * Read one char from input
 */
static void nextchar(void)
{
currch = fdgetc(fd);
if (currch == '\n')
	lineno++;
}


/* reservedword
 * Looks up a string in the reserved word table, rwtab.
 * This table must be defined in the parser into which this file is included.
 */
static int reservedword(char *s)
{
int i;

for(i=0; i<sizeof(rwtab)/sizeof(rwtab[0]); i++)
	if (strequ(rwtab[i].word,s))
		return rwtab[i].tok;
return -1;
}


/* yyerror - error reporting function
 */
void yyerror(char *s)
{
fprintf(stderr,"error : %s\nnear line %d\n",s,lineno);
}


/* yylex - special version for this parser, for handing double symbols
 * like >= and !=
 */
int yylex(void)
{
char tok[256];
char *tp = tok;

/* skip whitespace
 */
while(currch == ' ' || currch == '\t' || currch == '\n') {
	nextchar();
	}

if (isdigit(currch) || currch == '.') {
	/* A number
	 */
	double num;

	while(isdigit(currch) || currch == '.') {
		*tp++ = currch;
		nextchar();
		}
	*tp = '\0';
	yylval.num = atof(tok);
	if (parse_debug)
		printf("num : %f\n",yylval.num);
	return NUM;
	}
else if (currch == '\"') {
	/* A quoted string
	 */
	nextchar();
	while(currch != '\"') {
		if (currch == '\\') {
			/* escape char */
			nextchar();
			switch(currch) {
			case '\\': currch = '\\'; break;
			case 'n':  currch = '\n'; break;
			case 't':  currch = '\t'; break;
			case '\"': currch = '\"'; break;
			case '\'': currch = '\''; break;
				}
			}
		*tp++ = currch;
		nextchar();
		}
	*tp = '\0';
	nextchar();
	strcpy(yylval.string,tok);
	if (parse_debug)
		printf("string : %s\n",yylval.string);
	return STRING;
	}
else if (isalpha(currch) || currch == '_') {
	/* An alphanumeric string
	 */
	int rw;

	while(isalpha(currch) || isdigit(currch) || currch == '_') {
		*tp++ = currch;
		nextchar();
		}
	*tp = '\0';
	if ((rw = reservedword(tok)) != -1) {
		if (parse_debug)
			printf("reserved : %d = %s\n",rw,tok);
		return rw;
		}
	strcpy(yylval.string,tok);
	if (parse_debug)
		printf("string : %s\n",yylval.string);
	return WORD;
	}
else if (currch == '#') {
	/* cpp file info line
	 */
	char line[256];		/* line starting with '#' */
	char *lnum;		/* real line number */
	char *fnam;		/* real file name */
	int i=0;

	while(currch != '\n') {
		nextchar();
		line[i++] = currch;
		}
	lnum = strtok(line," \n");
	fnam = strtok(NULL," \n");
	lineno = atoi(lnum);
	nextchar();
	return yylex();		/* call again to get next token */
	}
else if (currch == -1) {
	/* End of file
	 */
	return END;
	}
else {
	/* Some other char
	 */
	int ch = currch;

	if (parse_debug)
		printf("char : %c\n",ch);
	nextchar();
	if (ch == '=' && currch == '=') {
		nextchar();
		return EQ;
		}
	else if (ch == '!' && currch == '=') {
		nextchar();
		return NEQ;
		}
	else if (ch == '<' && currch == '=') {
		nextchar();
		return LEQ;
		}
	else if (ch == '>' && currch == '=') {
		nextchar();
		return GEQ;
		}
	else if (ch == '+' && currch == '=') {
		nextchar();
		return PLUSEQ;
		}
	else if (ch == '-' && currch == '=') {
		nextchar();
		return MINUSEQ;
		}
	else if (ch == '*' && currch == '=') {
		nextchar();
		return TIMESEQ;
		}
	else if (ch == '/' && currch == '=') {
		nextchar();
		return DIVEQ;
		}
	else if (ch == '+' && currch == '+') {
		nextchar();
		return PLUSPLUS;
		}
	else if (ch == '-' && currch == '-') {
		nextchar();
		return MINUSMINUS;
		}
	else if (ch == '|' && currch == '|') {
		nextchar();
		return OR;
		}
	else if (ch == '&' && currch == '&') {
		nextchar();
		return AND;
		}
	return ch;
	}
}


/* parse_action
 * Read one action from a file descriptor. Returns 0 if ok, 1 on error or
 * -1 if 'end' was read.
 */
int parse_action(int filed, action **aptr)
{
int rv;

fd = filed;
currch = ' ';
lineno = 0;
rv = yyparse();
if (rv == 0) {
	*aptr = na;
	rv = init_vars(na,NULL,vst_static) == 1 ? 0 : 1;
	}
return rv;
}


/* parse_actions
 * Read actions till and end, and add them to the hash table
 */
int parse_actions(int filed)
{
int rv;

fd = filed;
currch = ' ';
lineno = 0;
while((rv = yyparse()) == 0) {
	if (init_vars(na,NULL,vst_static) == 0) {
		rv = 1;
		break;
		}
	add_action(na);
	}
if (rv == 1)
	return 1;
else
	return 0;
}


/* new_pnode
 * Allocate and perform some common initialisation of a parse node
 */
static pnode *new_pnode(int tp)
{
pnode *nn = create_pnode();

nn->type = tp;
nn->lineno = lineno;
return nn;
}


/* create_binop
 * Convenience function for allocating and filling a binary operator
 * parse node.
 */
static pnode *create_binop(int tp, pnode *l, pnode *r)
{
pnode *nn = new_pnode(tp);

nn->val.nbinop.l = l;
nn->val.nbinop.r = r;
return nn;
}


/* write_action
 * Output one action, its variables and its methods, to a file descriptor
 */
void write_action(int fd, int tl, action *a)
{
int i;

fdprintf(fd,"%saction \"%s\" {\n",tabs(tl),a->name);
for(i=4; i<a->varc; i++)
	write_variable(fd,tl+1,a->var[i]);
if (a->varc > 4)
	fdprintf(fd,"%s\n",tabs(tl));
for(i=0; i<a->metc; i++)
	write_method(fd,tl+1,a->met[i]);
fdprintf(fd,"%s}\n",tabs(tl+1));
}


/* write_variable
 * Output a variable as text to a fd
 */
static void write_variable(int fd, int tl, variable *v)
{
int i,len;
char type[20];

/* Create type name
 */
switch(v->type) {
case vt_void: sprintf(type,"void"); break;
case vt_dbl:  sprintf(type,"dbl"); break;
case vt_str:  sprintf(type,"str"); break;
case vt_var:  sprintf(type,"var"); break;
case vt_dblarray:
case vt_strarray:
	len = v->storage == vst_local ? v->val.arr.ni : v->val.arr.n;
	strcpy(type,v->type == vt_dblarray ? "dbl" : "str");
	if (len) sprintf(type+strlen(type),"[%d]",len);
	else     strcat(type,"[]");
	break;
	}

/* Output variable storage, type and name
 */
fdprintf(fd,"%s%s %s %s ",tabs(tl),v->storage==vst_local ? "local" : "static",
			  type,v->name);

if (v->storage == vst_local) {
	/* Output initialising expression
	 */
	if (v->init) {
		fdprintf(fd,"= ");
		write_ptree(fd,0,v->init);
		}
	}
else {
	/* Output current value
	 */
	switch(v->type) {
	case vt_void:
		break;
	case vt_dbl:
		fdprintf(fd,"= %f",v->val.dbl);
		break;
	case vt_str:
		fdprintf(fd,"=");
		len = strlen(v->val.str);
		if (len < 40)
			fdprintf(fd,"\"%s\"",expandstr(v->val.str));
		else {
			for(i=0; i<len; i+=40) {
				char temp[41];
				int j;

				if (i != 0) fdprintf(fd," &");
				for(j=0; j<40 && i+j<len; j++)
					temp[j] = v->val.str[i+j];
				temp[j] = 0;
				fdprintf(fd,"\n%s\"%s\"",tabs(tl+1),
							 expandstr(temp));
				}
			}
		break;
	case vt_var:
		fdprintf(fd,"= &%s",v->name,v->val.var->name);
		break;
	case vt_dblarray:
		if (v->val.arr.n) {
			fdprintf(fd,"= [ %f ",v->val.arr.a[0].val.dbl);
			for(i=1; i<v->val.arr.n; i++)
				fdprintf(fd,"; %f ",v->val.arr.a[i].val.dbl);
			fdprintf(fd,"]");
			}
		break;
	case vt_strarray:
		if (v->val.arr.n) {
			fdprintf(fd,"= [ \"%s\" ",v->val.arr.a[0].val.str);
			for(i=1; i<v->val.arr.n; i++)
				fdprintf(fd,"; \"%s\" ",
					 v->val.arr.a[i].val.str);
			fdprintf(fd,"]");
			}
		break;

		} /* end switch */
	}
fdprintf(fd,";\n");
}


/* write_method
 * Output one method, its variables and its code to a fd.
 */
static void write_method(int fd, int tl, method *m)
{
int i;

fdprintf(fd,"%smethod",tabs(tl));
for(i=0; i<m->namec; i++)
	fdprintf(fd," %s",m->name[i]);
fdprintf(fd," {\n");
for(i=0; i<m->varc; i++)
	write_variable(fd,tl+1,m->var[i]);
if (m->varc)
	fdprintf(fd,"%s\n",tabs(tl+1));
write_ptree(fd,tl+1,m->ptree);
fdprintf(fd,"%s}\n",tabs(tl+1));
}


/* write_ptree
 * Recursively output a parse tree as it was originally read from the source
 */
static void write_ptree(int fd, int tl, pnode *n)
{
pnode *pp;		/* for following lists */
char *bop[] = {"==","!=","+","-","*","/","%",">","<",">=","<=","||","&&"};
char *lvop[] = {"+=","-=","*=","/="};
char *unop[] = {"-","+","!"};

switch(n->type) {
/* Statements
 */
case pn_if:
	fdprintf(fd,"%sif (",tabs(tl));
	write_ptree(fd,tl+1,n->val.nif.cond);
	fdprintf(fd,")\n");
	write_ptree(fd,tl+1,n->val.nif.tr);
	if (n->val.nif.fl) {
		fdprintf(fd,"%selse\n",tabs(tl));
		write_ptree(fd,tl+1,n->val.nif.fl);
		}
	break;
case pn_block:
	fdprintf(fd,"%s{\n",tabs(tl));
	for(pp=n->val.nblock.st; pp; pp=pp->next)
		write_ptree(fd,tl,pp);
	fdprintf(fd,"%s}\n",tabs(tl));
	break;
case pn_cmds:
	for(pp=n->val.nblock.st; pp; pp=pp->next)
		write_ptree(fd,tl,pp);
	break;
case pn_while:
	fdprintf(fd,"%swhile (",tabs(tl));
	write_ptree(fd,tl+1,n->val.nwhile.cond);
	fdprintf(fd,")\n");
	write_ptree(fd,tl+1,n->val.nwhile.blk);
	break;
case pn_do:
	fdprintf(fd,"%sdo\n",tabs(tl));
	write_ptree(fd,tl+1,n->val.ndo.blk);
	fdprintf(fd,"%swhile (",tabs(tl));
	write_ptree(fd,tl+1,n->val.ndo.cond);
	fdprintf(fd,");\n");
	break;
case pn_for:
	fdprintf(fd,"%sfor(",tabs(tl));
	write_ptree(fd,tl+1,n->val.nfor.init);
	fdprintf(fd,"; ");
	write_ptree(fd,tl+1,n->val.nfor.cond);
	fdprintf(fd,"; ");
	write_ptree(fd,tl+1,n->val.nfor.inc);
	fdprintf(fd,")\n");
	write_ptree(fd,tl+1,n->val.nfor.blk);
	break;
case pn_estat:
	fdprintf(fd,"%s",tabs(tl));
	write_ptree(fd,tl+1,n->val.nestat.exp);
	fdprintf(fd,";\n");
	break;
case pn_null:
	fdprintf(fd,"%s;\n",tabs(tl));
/* Expressions
 */
case pn_equals: case pn_notequals: case pn_add:  case pn_subtract:
case pn_mult:   case pn_divide:    case pn_mod:  case pn_greater:
case pn_less:   case pn_grequals:  case pn_lsequals: case pn_and: case pn_or:
	fdprintf(fd,"("); write_ptree(fd,tl,n->val.nbinop.l); fdprintf(fd,")");
	fdprintf(fd," %s ",bop[n->type - pn_equals]);
	fdprintf(fd,"("); write_ptree(fd,tl,n->val.nbinop.r); fdprintf(fd,")");
	break;
case pn_var:
	write_ptree(fd,tl,n->val.nvar.var);
	break;
case pn_assign:
	write_ptree(fd,tl,n->val.nassign.lv);
	fdprintf(fd," = ");
	write_ptree(fd,tl,n->val.nassign.val);
	break;
case pn_const:
	if (n->val.nconst.n == (int)n->val.nconst.n)
		fdprintf(fd,"%d",(int)n->val.nconst.n);
	else
		fdprintf(fd,"%f",n->val.nconst.n);
	break;
case pn_addeq:  case pn_subtracteq:
case pn_multeq: case pn_divideeq:
	write_ptree(fd,tl,n->val.nlvalop.lv);
	fdprintf(fd," %s ",lvop[n->type - pn_addeq]);
	write_ptree(fd,tl,n->val.nlvalop.val);
	break;
case pn_unot: case pn_uminus: case pn_uplus:
	fdprintf(fd,"%s(",unop[n->type - pn_uminus]);
	write_ptree(fd,tl,n->val.nunop.r);
	fdprintf(fd,")");
	break;
case pn_concat:
	fdprintf(fd,"("); write_ptree(fd,tl,n->val.nconcat.l); fdprintf(fd,")");
	fdprintf(fd," & ");
	fdprintf(fd,"("); write_ptree(fd,tl,n->val.nconcat.r); fdprintf(fd,")");
	break;
case pn_sconst:
	fdprintf(fd,"\"%s\"",expandstr(n->val.nsconst.str));
	break;
/* Lvalues
 */
case pn_lval:
	fdprintf(fd,"%s",n->val.nlval.loc->name);
	break;
/* Function call
 */
case pn_fcall:
	fdprintf(fd,"%s(",n->val.nfcall.name);
	for(pp = n->val.nfcall.ar; pp; pp=pp->next) {
		write_ptree(fd,tl,pp);
		if (pp->next)
			fdprintf(fd,",");
		}
	fdprintf(fd,")");
	break;
/* Increment and decrement
 */
case pn_preinc:
	fdprintf(fd,"++"); write_ptree(fd,tl,n->val.nincdec.v);
	break;
case pn_postinc:
	write_ptree(fd,tl,n->val.nincdec.v); fdprintf(fd,"++");
	break;
case pn_predec:
	fdprintf(fd,"--"); write_ptree(fd,tl,n->val.nincdec.v);
	break;
case pn_postdec:
	write_ptree(fd,tl,n->val.nincdec.v); fdprintf(fd,"--");
	break;
/* Sizeof
 */
case pn_sizeof:
	fdprintf(fd,"sizeof(");
	write_ptree(fd,tl,n->val.nsizeof.a);
	fdprintf(fd,")");
	break;
/* Typeof
 */
case pn_typeof:
	fdprintf(fd,"typeof(");
	write_ptree(fd,tl,n->val.ntypeof.e);
	fdprintf(fd,")");
	break;
/* Array expressions
 */
case pn_aconst:
	fdprintf(fd,"[ ");
	for(pp=n->val.naconst.l; pp; pp=pp->next) {
		write_ptree(fd,tl,pp);
		if (pp->next)
			fdprintf(fd," ; ");
		}
	fdprintf(fd," ]");
	break;
case pn_aindex:
	write_ptree(fd,tl,n->val.naindex.arr);
	fdprintf(fd,"[");
	write_ptree(fd,tl,n->val.naindex.ind);
	fdprintf(fd,"]");
	break;
	} /* end switch */
}


/* write_actions
 * Output all actions to a file descriptor
 */
void write_actions(int fd)
{
action *a;

reset_action();
while((a = next_action())) {
	write_action(fd,0,a);
	fdputc(fd,'\n');
	}
fdprintf(fd,"end\n");
}

